#define MAXSECTORSOCCUPIED 20 

#include "cssysdef.h"
#include "qsqrt.h"
#include "ivaria/view.h"
#include "imesh/object.h"
#include "imesh/thing/thing.h"
#include "imesh/thing/polygon.h"
#include "iengine/mesh.h"
#include "igeom/polymesh.h"
#include "csgeom/transfrm.h"
#include "iengine/engine.h"
#include "iengine/sector.h"
#include "iengine/camera.h"
#include "iengine/region.h"
#include "imesh/terrfunc.h"
#include "iengine/movable.h"
#include "iutil/objreg.h"
#include "iutil/plugin.h"

#include "collisiondetection.h"
#include "phantom.h"
#include "demo1.h"

#undef GetObject

SCF_VERSION (TerrainInfo, 0, 0, 1);  // This is needed if the loaded map uses the terrain engine.
struct TerrainInfo : public csObject
{
	iTerrFuncState* terrfunc;
	SCF_DECLARE_IBASE_EXT (csObject);
};

SCF_IMPLEMENT_IBASE_EXT (TerrainInfo)
SCF_IMPLEMENTS_INTERFACE (TerrainInfo)
SCF_IMPLEMENT_IBASE_EXT_END

////////////////////////////////////// Collision Detection Class //////////////////////////////////////////////////

/** Constructor. */
CollisionDetection::CollisionDetection(Demo1* app): 
DX(0.4),
DY(1.0),
DZ(0.3),
OY(-0.9),
DX_L(0.3),
DZ_L(0.2),	
DX_2(DX/2),
DZ_2(DZ/2),
DX_2L(DX_L/2),
DZ_2L(DZ_L/2),
OYL(-1.7),
DYL(OY - OYL)

{
	do_gravity = true;
	move_3d = false;
	on_ground = true;
	this->app = app;
	do_cd = true;
	CreateColliders();
	InitCollDet(this->app->engine,NULL);
}

/** Destructor. */
CollisionDetection::~CollisionDetection() {

	delete app;
}

/** Calculates the line of collision between two triangles. */

int CollisionDetection::FindIntersection(csCollisionPair& cd,csVector3 line[2]) {

	csVector3 tri1[3]; tri1[0]=cd.a1; tri1[1]=cd.b1; tri1[2]=cd.c1;
	csVector3 tri2[3]; tri2[0]=cd.a2; tri2[1]=cd.b2; tri2[2]=cd.c2;
	
	return csMath3::FindIntersection(tri1,tri2,line);
}

/** Creates colliderWrappers for the avatar used to move around the world. */
void CollisionDetection::CreateColliders () {

	iPolygon3D *p;
	csRef<iPolygonMesh> mesh;
	csRef<iPluginManager> plugin_mgr (CS_QUERY_REGISTRY (app->object_reg, iPluginManager));
	
	// Get the rapid collide system plugin.
	const char* c = "crystalspace.collisiondetection.rapid";
	collide_system = CS_LOAD_PLUGIN (plugin_mgr, c, iCollideSystem);
	if (!collide_system)
	{
		printf("No Collision Detection plugin found!\n");
		exit(-1);
	}	
	
	csRef<iMeshObjectType> ThingType (CS_QUERY_PLUGIN_CLASS (plugin_mgr,"crystalspace.mesh.object.thing", iMeshObjectType));
	if (!ThingType)
		ThingType = CS_LOAD_PLUGIN (plugin_mgr,"crystalspace.mesh.object.thing", iMeshObjectType);
	
	csRef<iMeshObjectFactory> thing_fact (ThingType->NewFactory ());
	csRef<iMeshObject> mesh_obj (SCF_QUERY_INTERFACE (thing_fact, iMeshObject));
	plbody = app->engine->CreateMeshWrapper (mesh_obj, "Player's Body");
	csRef<iThingState> thing_state (SCF_QUERY_INTERFACE (mesh_obj, iThingState));
	
	thing_state->CreateVertex (csVector3 (-DX_2, OY,    -DZ_2));
	thing_state->CreateVertex (csVector3 (-DX_2, OY,    DZ_2));
	thing_state->CreateVertex (csVector3 (-DX_2, OY+DY, DZ_2));
	thing_state->CreateVertex (csVector3 (-DX_2, OY+DY, -DZ_2));
	thing_state->CreateVertex (csVector3 (DX_2,  OY,    -DZ_2));
	thing_state->CreateVertex (csVector3 (DX_2,  OY,    DZ_2));
	thing_state->CreateVertex (csVector3 (DX_2,  OY+DY, DZ_2));
	thing_state->CreateVertex (csVector3 (DX_2,  OY+DY, -DZ_2));
	
	// Left
	p = thing_state->CreatePolygon ();
	p->CreateVertex (0); p->CreateVertex (1);
	p->CreateVertex (2); p->CreateVertex (3);
	
	// Right
	p = thing_state->CreatePolygon ();
	p->CreateVertex (4); p->CreateVertex (5);
	p->CreateVertex (6); p->CreateVertex (7);
	
	// Bottom
	p = thing_state->CreatePolygon ();
	p->CreateVertex (0); p->CreateVertex (1);
	p->CreateVertex (5); p->CreateVertex (4);
	
	// Top
	p = thing_state->CreatePolygon ();
	p->CreateVertex (3); p->CreateVertex (2);
	p->CreateVertex (6); p->CreateVertex (7);
	
	// Front
	p = thing_state->CreatePolygon ();
	p->CreateVertex (1); p->CreateVertex (5);
	p->CreateVertex (6); p->CreateVertex (2);
	
	// Back
	p = thing_state->CreatePolygon ();
	p->CreateVertex (0); p->CreateVertex (4);
	p->CreateVertex (7); p->CreateVertex (3);
	
	mesh = SCF_QUERY_INTERFACE (mesh_obj, iPolygonMesh);
	body = new csColliderWrapper (plbody->QueryObject (), collide_system, mesh);
	body->SetName ("player body");
	plbody->GetRadius (body_radius, body_center);
	
	thing_fact = ThingType->NewFactory ();
	mesh_obj = SCF_QUERY_INTERFACE (thing_fact, iMeshObject);
	pllegs = app->engine->CreateMeshWrapper (mesh_obj, "Player's Legs");
	thing_state = SCF_QUERY_INTERFACE (mesh_obj, iThingState);
	
	thing_state->CreateVertex (csVector3 (-DX_2L, OYL,     -DZ_2L));
	thing_state->CreateVertex (csVector3 (-DX_2L, OYL,     DZ_2L));
	thing_state->CreateVertex (csVector3 (-DX_2L, OYL+DYL, DZ_2L));
	thing_state->CreateVertex (csVector3 (-DX_2L, OYL+DYL, -DZ_2L));
	thing_state->CreateVertex (csVector3 (DX_2L,  OYL,     -DZ_2L));
	thing_state->CreateVertex (csVector3 (DX_2L,  OYL,     DZ_2L));
	thing_state->CreateVertex (csVector3 (DX_2L,  OYL+DYL, DZ_2L));
	thing_state->CreateVertex (csVector3 (DX_2L,  OYL+DYL, -DZ_2L));
	
	// Left
	p = thing_state->CreatePolygon ();
	p->CreateVertex (0); p->CreateVertex (1);
	p->CreateVertex (2); p->CreateVertex (3);
	
	// Right
	p = thing_state->CreatePolygon ();
	p->CreateVertex (4); p->CreateVertex (5);
	p->CreateVertex (6); p->CreateVertex (7);
	
	// Bottom
	p = thing_state->CreatePolygon ();
	p->CreateVertex (0); p->CreateVertex (1);
	p->CreateVertex (5); p->CreateVertex (4);
	
	// Top
	p = thing_state->CreatePolygon ();
	p->CreateVertex (3); p->CreateVertex (2);
	p->CreateVertex (6); p->CreateVertex (7);
	
	// Front
	p = thing_state->CreatePolygon ();
	p->CreateVertex (1); p->CreateVertex (5);
	p->CreateVertex (6); p->CreateVertex (2);
	
	// Back
	p = thing_state->CreatePolygon ();
	p->CreateVertex (0); p->CreateVertex (4);
	p->CreateVertex (7); p->CreateVertex (3);
	
	mesh = SCF_QUERY_INTERFACE (mesh_obj, iPolygonMesh);
	legs = new csColliderWrapper (pllegs->QueryObject (), collide_system, mesh);
	legs->SetName ("player legs");
	pllegs->GetRadius ( legs_radius, legs_center);
	
	SCF_DEC_REF (legs);
	SCF_DEC_REF (body);
	
	if (!body || !legs)
		do_cd = false;
}

/** Finds number of sectors that are nearby a point. */
int CollisionDetection::FindSectors (csVector3 v, csVector3 d, iSector *s, iSector **sa) {

	int c = 0;
	// @@@ Avoid this sqrt somehow? i.e. by having it in the objects.
	float size = qsqrt (d.x * d.x + d.y * d.y + d.z * d.z);
	csRef<iSectorIterator> it (app->engine->GetNearbySectors (s, v, size));
	iSector* sector;
	while ((sector = it->Fetch ()) != NULL)
	{
		sa[c++] = sector;
		if (c >= MAXSECTORSOCCUPIED) break;
	}
	return c;
}

/**	Tests collision detection for a csColliderwrapper and nearby objects within a certain radius in the same sector. 
*	Returns number of collisions.
*/

int CollisionDetection::CollisionDetect (iEngine* Engine, csColliderWrapper *c, iSector* sp, csReversibleTransform *cdt) {

	int hit = 0;
	int j;

	// Check collision with this sector.
	csCollisionPair* CD_contact;
	
	csRef<iObjectIterator> objitaaa (Engine->GetNearbyObjects (sp, cdt->GetOrigin (), 3));		// 3 should be enough for moving around.
	while (!objitaaa->IsFinished ()) {

		iObject* mw_obj = objitaaa->GetObject();
		csRef<iMeshWrapper> mw (SCF_QUERY_INTERFACE (mw_obj, iMeshWrapper));
		if (mw)
		{
			collide_system->ResetCollisionPairs ();
			if (c->Collide (mw_obj, cdt, &mw->GetMovable ()->GetTransform ())) hit++;
			
			CD_contact = collide_system->GetCollisionPairs ();
			for (j=0 ; j<collide_system->GetCollisionPairCount () ; j++)
				our_cd_contact[num_our_cd++] = CD_contact[j];
			
			if (collide_system->GetOneHitOnly () && hit)
				return 1;
			// TODO, should test which one is the closest.
		}
		objitaaa->Next ();
	}
	
	return hit;
}

/** Create colliderWrappers for all geometry in the world */ 
void CollisionDetection::InitCollDet (iEngine* engine, iRegion* region) {
	
    csRef<iPolygonMesh> mesh;
    // Initialize all mesh objects for collision detection.
    int i;
    iMeshList* meshes = engine->GetMeshes ();
    for (i = 0 ; i < meshes->GetCount () ; i++) {
		iMeshWrapper* sp = meshes->Get (i);
		if (region && !region->IsInRegion (sp->QueryObject ())) continue;
		mesh = SCF_QUERY_INTERFACE (sp->GetMeshObject (), iPolygonMesh);
		if (mesh) {
			csColliderWrapper *cw = new csColliderWrapper (sp->QueryObject (), collide_system, mesh);
			cw->SetName (sp->QueryObject ()->GetName());
			cw->DecRef ();
		}
    }
}

/** Test collision detection using both gravity and checking for collision when moving around in the world */
void CollisionDetection::DoGravity (iEngine* Engine, csVector3& pos, csVector3& vel) {

	pos=app->view->GetCamera ()->GetTransform ().GetOrigin ();
	
	csVector3 new_pos = pos+vel;
	csMatrix3 m;
	csOrthoTransform test (m, new_pos);
	
	iSector* n[MAXSECTORSOCCUPIED];
	int num_sectors = FindSectors (new_pos, 4.0f*body_radius,
		app->view->GetCamera()->GetSector(), n);
	
	num_our_cd = 0;
	collide_system->SetOneHitOnly (false);
	int hits = 0;
	
	// Check to see if there are any terrains, if so test against those.
	// This routine will automatically adjust the transform to the highest
	// terrain at this point.
	
	// @@@@@@ The following code supports only one terrain in a sector!
	int k;
	for ( k = 0; k < num_sectors ; k++)	// This is needed if the loaded map uses the terrain engine.
	{
		iMeshList* ml = n[k]->GetMeshes ();
		if (ml->GetCount () > 0)
		{
			csRef<TerrainInfo> ti (CS_GET_CHILD_OBJECT (n[k]->QueryObject (),
				TerrainInfo));
			if (ti)
			{
				if (ti->terrfunc)
				{
					hits += ti->terrfunc->CollisionDetect (&test);
				}
			}
			else
			{
				ti = csPtr<TerrainInfo> (new TerrainInfo ());
				ti->terrfunc = NULL;	// No terrain found yet.
				int i;
				for (i = 0 ; i < ml->GetCount () ; i++)
				{
					iMeshWrapper* terrain = ml->Get (i);
					csRef<iTerrFuncState> state (SCF_QUERY_INTERFACE (terrain
						->GetMeshObject (), iTerrFuncState));
					if (state)
					{
						hits += state->CollisionDetect (&test);
						ti->terrfunc = state;
						break;
					}
				}
				csRef<iObject> iobj (SCF_QUERY_INTERFACE (ti, iObject));
				n[k]->QueryObject ()->ObjAdd (iobj);
			}
		}
	}
	
	// If there were hits with the terrain we update our new position
	// here. Side note: this could moved outside the loop above because
	// a compiler bug with gcc 2.7.2 prevented it from working when inside
	// the loop.
	if (hits) 
		new_pos = test.GetOrigin ();
	
	int j;
	
	if (hits == 0)
	{
		collide_system->ResetCollisionPairs ();
		
		for ( ; num_sectors-- ; )
			hits += CollisionDetect (Engine, body, n[num_sectors], &test); //num_our_cd is calculated
		
		for (j=0 ; j<num_our_cd; j++)
		{
			csCollisionPair& cd = our_cd_contact[j];
			csVector3 n = ((cd.c2-cd.b2)%(cd.b2-cd.a2)).Unit(); // calculate normal
			if (n*vel<0)
				continue;
			vel = -(vel%n)%n;
		}
		
		// We now know our (possible) velocity. Let's try to move up or down, if possible
		new_pos = pos+vel;

		// Try again, and don't move if we're still in a wall
		int num_sectors = FindSectors (new_pos, 4.0f*body_radius, app->view->GetCamera()->GetSector(), n);
		collide_system->SetOneHitOnly (false);
		collide_system->ResetCollisionPairs ();
		test = csOrthoTransform (csMatrix3(), new_pos);
		int hit = 0;
		for (; num_sectors--;) {
			if (CollisionDetect (Engine, body, n[num_sectors], &test) > 0)
			{
				new_pos-=vel;
				break;
			}
		}
		
		test = csOrthoTransform (csMatrix3(), new_pos);
		
		num_sectors = FindSectors (new_pos, 4.0f*legs_radius, app->view->GetCamera()->GetSector(), n);
		
		num_our_cd = 0;
		collide_system->SetOneHitOnly (false);
		collide_system->ResetCollisionPairs ();
		
		for ( ; num_sectors-- ; ) {
			hit += CollisionDetect (Engine, legs, n[num_sectors], &test);
		}
		
		if (!hit)
		{
			on_ground = false;
			if (do_gravity && !move_3d)
				vel.y -= 0.002;
		}
		else
		{
			float max_y=-1e10;
			
			for (j=0 ; j<num_our_cd ; j++)
			{
				csCollisionPair cd = our_cd_contact[j];
				csVector3 n = ((cd.c2-cd.b2)%(cd.b2-cd.a2)).Unit();
				
				if (n*csVector3(0,-1,0)<0.7) continue;
				
				csVector3 line[2];
				
				cd.a1 += new_pos;
				cd.b1 += new_pos;
				cd.c1 += new_pos;
				
				if (FindIntersection (cd,line))
				{
					if (line[0].y>max_y)
						max_y=line[0].y;
					if (line[1].y>max_y)
						max_y=line[1].y;
				}
			}
			
			float p = new_pos.y-max_y+OYL+0.01;
			if (ABS(p)<DYL-0.01)
			{
				if (max_y != -1e10)
					new_pos.y = max_y-OYL-0.01;
				
				if (vel.y<0)
					vel.y = 0;
			}
			on_ground = true;
		}
	}

	app->setWorkspacePosition(new_pos.x, new_pos.y, new_pos.z);	

	new_pos -= app->view->GetCamera ()->GetTransform ().GetOrigin (); // Relative position 
	app->view->GetCamera ()->MoveWorld (new_pos);
	app->velocity = app->view->GetCamera ()->GetTransform ().GetO2T ()*vel;
	app->velocity.x = 0;
	
	if(!do_gravity)
		app->velocity.y -= SIGN (app->velocity.y) * MIN (0.017, ABS (app->velocity.y));	
}
