#include "cssysdef.h"

#include "cssys/sysfunc.h"
#include "iutil/vfs.h"
#include "csutil/cscolor.h"
#include "cstool/csview.h"
#include "cstool/initapp.h"
#include "iutil/eventq.h"
#include "iutil/event.h"
#include "iutil/objreg.h"
#include "iutil/object.h"
#include "iutil/csinput.h"
#include "iutil/virtclk.h"
#include "iengine/sector.h"
#include "iengine/engine.h"
#include "iengine/camera.h"
#include "iengine/light.h"
#include "iengine/statlght.h"
#include "iengine/texture.h"
#include "iengine/mesh.h"
#include "iengine/movable.h"
#include "iengine/material.h"
#include "iengine/campos.h"
#include "imesh/thing/polygon.h"
#include "imesh/thing/thing.h"
#include "imesh/object.h"
#include "imesh/sprite3d.h"
#include "imesh/ball.h"
#include "ivideo/graph3d.h"
#include "ivideo/graph2d.h"
#include "ivideo/txtmgr.h"
#include "ivideo/texture.h"
#include "ivideo/material.h"
#include "ivideo/fontserv.h"
#include "igraphic/imageio.h"
#include "imap/parser.h"
#include "ivaria/reporter.h"
#include "ivaria/stdrep.h"
#include "csutil/cmdhelp.h"
#include "csutil/cmdline.h"
#include "csutil/scf.h"
#include "csutil/cfgfile.h"

#include "collisiondetection.h"
#include "phantom.h"
#include "demo1.h"

CS_IMPLEMENT_APPLICATION

// For the case we're using static linking ...
#ifdef CS_STATIC_LINKED
SCF_REGISTER_STATIC_LIBRARY (phantom)
#endif

Demo1* app;

/** Constructor. */
Demo1::Demo1(iObjectRegistry* obj_reg) {
	tiltAngle = 0;
	object_reg = obj_reg;
	validConstruction = false;
	velocity.x=0;velocity.y=0;velocity.z=0;
	forcesOn = true;
	pressed = false;
	started = false;
	frames = 0;
	startTime = 0;
}

/** Destuctor. */
Demo1::~Demo1() {
	delete colldet;
}

/** Initializes the plugins, loads the map and prepares the engine. */
bool Demo1::Initialize (int argc, char *argv[]) {
	
	csRef<iCommandLineParser> cmdline (CS_QUERY_REGISTRY (object_reg, iCommandLineParser));
	const char *val;
	map_dir = new char[512];
	if ((val = cmdline->GetName ())){
		// if an absolute path is given, copy it. Otherwise prepend "/lev/".
		if (val[0] == '/')
			strcpy (map_dir, val);
		else
			sprintf (map_dir, "/lev/%s", val);
	}
	else
		strcpy(map_dir,"/lev/bighouse");	
	
	if (!csInitializer::RequestPlugins (object_reg,
		CS_REQUEST_VFS,
		CS_REQUEST_SOFTWARE3D,
		CS_REQUEST_ENGINE,
		CS_REQUEST_FONTSERVER,
		CS_REQUEST_IMAGELOADER,
		CS_REQUEST_LEVELLOADER,
		CS_REQUEST_REPORTER,
		CS_REQUEST_PLUGIN("crystalspace.plugins.phantom",iPhantom),
		CS_REQUEST_REPORTERLISTENER,
		CS_REQUEST_END))
	{
		csReport (object_reg, CS_REPORTER_SEVERITY_ERROR,
			"crystalspace.application.demo1",
			"Can't initialize plugins!");
		return false;
	}
	// Setup the EventHandler.
	if (!csInitializer::SetupEventHandler (object_reg, SimpleEventHandler))
	{
		csReport (object_reg, CS_REPORTER_SEVERITY_ERROR,
			"crystalspace.application.demo1",
			"Can't initialize event handler!");
		return false;
	}
	
	// Check for commandline help.
	if (csCommandLineHelper::CheckHelp (object_reg))
	{
		csCommandLineHelper::Help (object_reg);
		return false;
	}
	
	// The virtual clock.
	vc = CS_QUERY_REGISTRY (object_reg, iVirtualClock);
	if (vc == NULL)
	{
		csReport (object_reg, CS_REPORTER_SEVERITY_ERROR,
			"crystalspace.application.demo1",
			"Can't find the virtual clock!");
		return false;
	}
	
	// Find the pointer to engine plugin.
	engine = CS_QUERY_REGISTRY (object_reg, iEngine);
	if (engine == NULL)
	{
		csReport (object_reg, CS_REPORTER_SEVERITY_ERROR,
			"crystalspace.application.demo1",
			"No iEngine plugin!");
		return false;
	}
	
	// Get the loader plugin.
	loader = CS_QUERY_REGISTRY (object_reg, iLoader);
	if (loader == NULL)
	{
		csReport (object_reg, CS_REPORTER_SEVERITY_ERROR,
			"crystalspace.application.demo1",
			"No iLoader plugin!");
		return false;
	}
	
	// The Graphics 3D plugin.
	g3d = CS_QUERY_REGISTRY (object_reg, iGraphics3D);
	if (g3d == NULL)
	{
		csReport (object_reg, CS_REPORTER_SEVERITY_ERROR,
			"crystalspace.application.demo1",
			"No iGraphics3D plugin!");
		return false;
	}
	
	// The Keyboard driver.
	kbd = CS_QUERY_REGISTRY (object_reg, iKeyboardDriver);
	if (kbd == NULL)
	{
		csReport (object_reg, CS_REPORTER_SEVERITY_ERROR,
			"crystalspace.application.demo1",
			"No iKeyboardDriver plugin!");
		return false;
	}
	
	// The Phantom plugin.
	phantom = CS_QUERY_REGISTRY (object_reg, iPhantom);
	if (!phantom) {
		fprintf (stderr, "No csPhantom shared class!\n");
	}
	
	// Open the main system. This will open all the previously loaded plug-ins.
	if (!csInitializer::OpenApplication (object_reg))
	{
		csReport (object_reg, CS_REPORTER_SEVERITY_ERROR,
			"crystalspace.application.demo1",
			"Error opening system!");
		return false;
	}

	// Check for a valid phantom construction.
	validConstruction = phantom->getValidConstruction();
		
	// Create a new camera view.
	view = csPtr<iView> (new csView (engine, g3d)); 
	g2d = g3d->GetDriver2D ();
	view->SetRectangle (0, 0, g2d->GetWidth (), g2d->GetHeight ());
	
	// Load the map that has been read from the command line.
	LoadMap();

	// Get the ball mesh object plug-in used for the graphic representation of the PHANTOM position.
	if (!loader->LoadTexture ("white", "/lib/std/green.gif"))
	{
		csReport (object_reg, CS_REPORTER_SEVERITY_ERROR,
			"crystalspace.application.demosky2",
			"Error loading 'white' texture!");
		exit (1);
	}
	const char* classId = "crystalspace.mesh.object.ball";
	csRef<iMeshFactoryWrapper> mesh_fact (engine->CreateMeshFactory(classId,
		"ballFact"));
	csVector3 meshposition(0,0,0);
	ball = engine->CreateMeshWrapper(mesh_fact,
		"PHANTOM", room, meshposition);
	csVector3 meshradius(0.05,0.05,0.05);
	csRef<iBallState> ballstate (SCF_QUERY_INTERFACE( ball->GetMeshObject(),
		iBallState));

	l_mat = (engine->CreateBaseMaterial (NULL, 0, NULL, NULL));
	l_mat->SetFlatColor (csRGBcolor (0, 255, 0));
	mw = engine->GetMaterialList ()->NewMaterial (l_mat);
	mw->QueryObject ()->SetName ("color");
	mw->Register (g3d->GetTextureManager ());
	mw->GetMaterialHandle ()->Prepare ();
	ballstate->SetRadius( meshradius.x, meshradius.y, meshradius.z );
	ballstate->SetRimVertices( 10 );
	ballstate->SetMaterialWrapper( mw );
	ballstate->SetLighting(true);
	ball->DeferUpdateLighting (CS_NLIGHT_STATIC|CS_NLIGHT_DYNAMIC, 10);

	engine->Prepare ();

	// Create a haptic representation.
	phantom->setMeshList(engine->GetMeshes());

	// Create the collision handling class.
	colldet = new CollisionDetection(this);
	
	return true;
}

/** Loads a map and sets the correct camera position in CS and the correct workspace position in the plugin. */
bool Demo1::LoadMap() {

	// Set VFS current directory to the level we want to load.
	csRef<iVFS> VFS (CS_QUERY_REGISTRY (object_reg, iVFS));
	VFS->ChDir (map_dir);
	// Load the level file which is called 'world'.
	if (!loader->LoadMapFile ("world"))
	{
		csReport (object_reg, CS_REPORTER_SEVERITY_ERROR,
			"crystalspace.application.simple",
			"Couldn't load level!");
		return false;
	}
	// Find the starting position in this level.
	csVector3 pos (0, 1, 0);
	if (engine->GetCameraPositions ()->GetCount () > 0)
	{
		// There is a valid starting position defined in the level file.
		iCameraPosition* campos = engine->GetCameraPositions ()->Get (0);
		room = engine->GetSectors ()->FindByName (campos->GetSector ());
		pos = campos->GetPosition ();
		if (validConstruction) phantom->setWorkspacePosition(pos.x, pos.y, pos.z);
	}
	else
	{
		// We didn't find a valid starting position. So we default
		// to going to room called 'room' at position (0,0,0).
		room = engine->GetSectors ()->FindByName ("room");
		if (room != NULL)
			printf("lr finnas ett room\n");
	}
	if (!room)
	{
		csReport (object_reg, CS_REPORTER_SEVERITY_ERROR,
			"crystalspace.application.simple",
			"Can't find a valid starting position!");
		return false;
	}
	view = csPtr<iView> (new csView (engine, g3d));
	view->GetCamera ()->GetTransform ().SetOrigin (pos);
	view->GetCamera ()->SetSector (room);
	
	return true;	
}

/**	Handles keyboard events when no phantom is used. */
void Demo1::updateKeyboardFrame() {

	// First get elapsed time from the virtual clock.
	csTicks elapsed_time = vc->GetElapsedTicks ();
	
	speed = (float)((elapsed_time / 1000.0) * (0.03 * 20));//20);
	
	iCamera* c = view->GetCamera();
	
	// Rotate right.
	if (kbd->GetKeyState (CSKEY_RIGHT)){
		c->GetTransform ().RotateOther (CS_VEC_ROT_RIGHT, speed*2);
	}
	// Rotate left.
	if (kbd->GetKeyState (CSKEY_LEFT)){
		c->GetTransform ().RotateOther (CS_VEC_ROT_LEFT, speed*2);
	}
	// Tilt up.
	if (kbd->GetKeyState (CSKEY_PGUP)){
		c->GetTransform ().RotateThis (CS_VEC_TILT_UP, speed*2);
	}
	// Tilt down.
	if (kbd->GetKeyState (CSKEY_PGDN)){
		c->GetTransform ().RotateThis (CS_VEC_TILT_DOWN, speed*2);
	}
	// Move forward.
	if (kbd->GetKeyState (CSKEY_UP)){
		if (velocity.z>0.02)
			velocity.z -= speed;
		else
			velocity.z += speed;
		pressed = true;
		
	}
	// Move backward.
	if (kbd->GetKeyState (CSKEY_DOWN)){
		if (velocity.z<-0.02)
			velocity.z += speed;
		else
			velocity.z -= speed;
		pressed = true;
	}
}

/**	Handles keyboard events and phantom events when a phantom is used. */
void Demo1::updatePhantomFrame() {
	
	// First get elapsed time from the virtual clock.
	csTicks elapsed_time = vc->GetElapsedTicks ();
	
	// Now rotate the camera according to keyboard state
	speed = (float)((elapsed_time / 1000.0) * (0.03 * 110));
	rotateSpeed = speed/4;
	rotate = csVector3(0);
	
	iCamera* c = view->GetCamera();

	// Rotate right. Rotate phantom workspace right.
	if (kbd->GetKeyState (CSKEY_RIGHT)) {
		c->GetTransform ().RotateOther (CS_VEC_ROT_RIGHT, rotateSpeed);
		rotate = csVector3(CS_VEC_ROT_RIGHT);
	}
	// Rotate left. Rotate phantom workspace left.
	if (kbd->GetKeyState (CSKEY_LEFT)) {
		c->GetTransform ().RotateOther (CS_VEC_ROT_LEFT, rotateSpeed);
		rotate = csVector3(CS_VEC_ROT_LEFT);
	}
	// Tilt up. Tilt phantom workspace up. 
	if (kbd->GetKeyState (CSKEY_PGUP)) {
		tiltAngle += rotateSpeed;
		if(tiltAngle <= 0.7) {
			c->GetTransform ().RotateThis (CS_VEC_TILT_UP, rotateSpeed);
		}
		else
			tiltAngle -= rotateSpeed;
	}
	// Tilt down. Tilt phantom workspace down. 
	if (kbd->GetKeyState (CSKEY_PGDN)) {
		tiltAngle -= rotateSpeed;
		if(tiltAngle >= -0.7) {
			c->GetTransform ().RotateThis (CS_VEC_TILT_DOWN, rotateSpeed);
		}
		else
			tiltAngle += rotateSpeed;
	}
	// Move forward. Move phantom workspace forward.
	if (kbd->GetKeyState (CSKEY_UP)) {
		csVector3 pos = c->GetTransform().GetOrigin();
		if (velocity.z>0.03)
			velocity.z -= speed;
		else
			velocity.z += speed;
		pressed = true;
	}
	// Move backward. Move phantom workspace forward.
	if (kbd->GetKeyState (CSKEY_DOWN)) {
		csVector3 pos = c->GetTransform().GetOrigin();
		if (velocity.z<-0.03)
			velocity.z += speed;
		else
			velocity.z -= speed;
		pressed = true;
	}
	
	// Check collision.If collision change the cursor color to red. 
	if(phantom->isCollision())
		l_mat->SetFlatColor(csRGBcolor (255, 0, 50));
	else												
		l_mat->SetFlatColor(csRGBcolor (0, 255, 50));
	
	mw->Register (g3d->GetTextureManager ());
	mw->GetMaterialHandle ()->Prepare ();
	
	// Update cursor position.
	ball->GetMovable()->SetPosition(phantom->getStylusPosition());
	ball->GetMovable()->UpdateMove();
}

/** Updates the frame. */
void Demo1::SetupFrame () {

	pressed = false;
	if (validConstruction)
		updatePhantomFrame();
	else
		updateKeyboardFrame();
	if (!pressed) {
		if (velocity.z > 0.01)
			velocity.z -= 0.01;
		else if (velocity.z < -0.01)
			velocity.z += 0.01;
		else
			velocity.z = 0;
	}
	
	// Tell 3D driver we're going to display 3D things.
	if (!g3d->BeginDraw (engine->GetBeginDrawFlags () | CSDRAW_3DGRAPHICS))
		return;
	
	csVector3 vel = view->GetCamera ()->GetTransform ().GetT2O ()*velocity;
	csVector3 pos;
	
	// Do collisiondetection ten times per frame to get a smoother servoloop update when moving the workspace.
	vel = vel/10;
	csVector3 move = newPosition - lastPosition;
	move = move/10;
	lastPosition = newPosition;
	int i = 0;
	while (i<10) {
		newPosition+=move;
		colldet->DoGravity(engine,pos,vel);
		phantom->rotateWorkspace(rotateSpeed/10,rotate.x,rotate.y,rotate.z);
		phantom->setWorkspacePosition(newPosition.x,newPosition.y,newPosition.z);
		i++;
	}
	view->Draw ();
}

/** Finishes the drawing of a frame. */
void Demo1::FinishFrame () {

	g3d->FinishDraw ();
	g3d->Print (NULL);
}

/** Handles CS events. */
bool Demo1::HandleEvent (iEvent& ev) {

	if (ev.Type == csevBroadcast && ev.Command.Code == cscmdProcess)
	{
		SetupFrame ();
		return true;
	}
	else if (ev.Type == csevBroadcast && ev.Command.Code == cscmdFinalProcess)
	{
		FinishFrame ();
		return true;
	}	
	else if (ev.Type == csevKeyDown && ev.Key.Code == 'f')
	{
		forcesOn = !forcesOn;
		phantom->setForcesOn(forcesOn);
		return true;
	}
	else if (ev.Type == csevKeyDown && ev.Key.Code == CSKEY_ESC)
	{
		csRef<iEventQueue> q (CS_QUERY_REGISTRY (object_reg, iEventQueue));
		phantom->stop();
		if (q) q->GetEventOutlet()->Broadcast (cscmdQuit);
		return true;
	}
	
	return false;
}

/** Set position of the phantom workspace. */
void Demo1::setWorkspacePosition(double x,double y,double z) {

	newPosition = csVector3(x,y,z);
}

/** Handles CS events. Needs to be static. Is registered to the event queue. Calls HandleEvent. */
bool Demo1::SimpleEventHandler (iEvent& ev) {

	return app->HandleEvent (ev);
}

/** Starts the default run loop and the plugin servo loop. */
void Demo1::Start () {

	if (validConstruction)
		phantom->start();
	csDefaultRunLoop (object_reg);
}

/** Main method. */
int main(int argc, char *argv[]) {
	
	iObjectRegistry* object_reg = csInitializer::CreateEnvironment (argc, argv);	
	app = new Demo1(object_reg); 
		
	app->Initialize(argc, argv);
	app->Start();
	app->phantom->DecRef();
	
	csInitializer::DestroyApplication (object_reg);

	return 0;	
}

