#include <stdio.h>

#include "cssysdef.h"
#include "cssys/sysfunc.h"
#include "iutil/vfs.h"
#include "csutil/cscolor.h"
#include "cstool/csview.h"
#include "cstool/initapp.h"
#include "cstool/cspixmap.h"
#include "iutil/eventq.h"
#include "iutil/event.h"
#include "iutil/objreg.h"
#include "iutil/csinput.h"
#include "iutil/virtclk.h"
#include "iengine/sector.h"
#include "iengine/engine.h"
#include "iengine/camera.h"
#include "iengine/light.h"
#include "iengine/statlght.h"
#include "iengine/texture.h"
#include "iengine/mesh.h"
#include "iengine/movable.h"
#include "iengine/material.h"
#include "iengine/campos.h"
#include "imesh/thing/polygon.h"
#include "imesh/thing/thing.h"
#include "imesh/object.h"
#include "imesh/sprite3d.h"
#include "imesh/ball.h"
#include "ivideo/graph3d.h"
#include "ivideo/graph2d.h"
#include "ivideo/txtmgr.h"
#include "ivideo/texture.h"
#include "ivideo/material.h"
#include "ivideo/fontserv.h"
#include "igraphic/imageio.h"
#include "imap/parser.h"
#include "ivaria/reporter.h"
#include "ivaria/stdrep.h"
#include "csutil/cmdhelp.h"
#include "csutil/cmdline.h"
#include "csutil/scf.h"
#include "csutil/cfgfile.h"
#include "ivaria/dynamics.h"
#include "igeom/objmodel.h"

#include "phantom.h"
#include "demo2.h"

CS_IMPLEMENT_APPLICATION

// for the case we're using static linking ...
#ifdef CS_STATIC_LINKED
SCF_REGISTER_STATIC_LIBRARY (phantom)
#endif

double CAMERA_POSITION_Y = 8;
const double PHANTOM_POSITION_Y = 0.2;
const float OFFSET = 1;
const float MAX_CAMERA_HEIGHT = 8;
const float MIN_CAMERA_HEIGHT = 4;

Demo2* pt;

/** Constructor. */
Demo2::Demo2(iObjectRegistry* obj_reg) {
	object_reg = obj_reg;
	useMap = true;
	step = 0.004;
	posX = 0.0005;
	validConstruction = false;
	firstTimeSwitch = true;
	stylusDistance = 0.0;
	minX = DBL_MAX;
	maxX = DBL_MIN;
	minZ = DBL_MAX;
	maxZ = DBL_MIN;
	firstTime = true;
	mode = 0;
	frames = 0;
	workspaceRotated = false;
	cmdLevel = false;
}

/** Destructor. */
Demo2::~Demo2() {
}

/** Initializes the plugins and loads the 2d images. */
bool Demo2::Initialize () {

	csRef<iCommandLineParser> cmdline (CS_QUERY_REGISTRY (object_reg, iCommandLineParser));
	const char *val;
	map_dir = new char[512];
	if ((val = cmdline->GetName ())){
		// if an absolute path is given, copy it. Otherwise prepend "/lev/".
		if (val[0] == '/')
			strcpy (map_dir, val);
		else
			sprintf (map_dir, "/lev/%s", val);
	mode = 3;
	cmdLevel = true;
	}
	else
		strcpy(map_dir,"/lev/labyrinth1");
	
	if (!csInitializer::RequestPlugins (object_reg,
		CS_REQUEST_VFS,
		CS_REQUEST_SOFTWARE3D,
		CS_REQUEST_ENGINE,
		CS_REQUEST_FONTSERVER,
		CS_REQUEST_IMAGELOADER,
		CS_REQUEST_LEVELLOADER,
		CS_REQUEST_REPORTER,
		CS_REQUEST_PLUGIN ("crystalspace.dynamics.ode", iDynamics),
		CS_REQUEST_PLUGIN("crystalspace.plugins.phantom",iPhantom),
		CS_REQUEST_REPORTERLISTENER,
		CS_REQUEST_END))
	{
		csReport (object_reg, CS_REPORTER_SEVERITY_ERROR,
			"crystalspace.application.demo2",
			"Can't initialize plugins!");
		return false;
	}
	if (!csInitializer::SetupEventHandler (object_reg, SimpleEventHandler))
	{
		csReport (object_reg, CS_REPORTER_SEVERITY_ERROR,
			"crystalspace.application.demo2",
			"Can't initialize event handler!");
		return false;
	}
	
	// Check for commandline help.
	if (csCommandLineHelper::CheckHelp (object_reg))
	{
		csCommandLineHelper::Help (object_reg);
		return false;
	}
	
	// The virtual clock.
	vc = CS_QUERY_REGISTRY (object_reg, iVirtualClock);
	if (vc == NULL)
	{
		csReport (object_reg, CS_REPORTER_SEVERITY_ERROR,
			"crystalspace.application.demo2",
			"Can't find the virtual clock!");
		return false;
	}
	
	// Find the pointer to engine plugin
	engine = CS_QUERY_REGISTRY (object_reg, iEngine);
	if (engine == NULL)
	{
		csReport (object_reg, CS_REPORTER_SEVERITY_ERROR,
			"crystalspace.application.demo2",
			"No iEngine plugin!");
		return false;
	}
	
	// Get the loader plugin.
	loader = CS_QUERY_REGISTRY (object_reg, iLoader);
	if (loader == NULL)
	{
		csReport (object_reg, CS_REPORTER_SEVERITY_ERROR,
			"crystalspace.application.demo2",
			"No iLoader plugin!");
		return false;
	}
	
	// Get the g3d plugin.
	g3d = CS_QUERY_REGISTRY (object_reg, iGraphics3D);
	if (g3d == NULL)
	{
		csReport (object_reg, CS_REPORTER_SEVERITY_ERROR,
			"crystalspace.application.demo2",
			"No iGraphics3D plugin!");
		return false;
	}
	
	// Get the g2d plugin.
	g2d = CS_QUERY_REGISTRY (object_reg, iGraphics2D);
	if (!g2d)
	{
		csReport (object_reg, CS_REPORTER_SEVERITY_ERROR, 
			"crystalspace.application.demo2",
			"No iGraphics2D plugin!");
		return false;
	}
	
	// Get the keyboard plugin.
	kbd = CS_QUERY_REGISTRY (object_reg, iKeyboardDriver);
	if (kbd == NULL)
	{
		csReport (object_reg, CS_REPORTER_SEVERITY_ERROR,
			"crystalspace.application.demo2",
			"No iKeyboardDriver plugin!");
		return false;
	}
	
	// Get the dynamics plugin.
	dyn = CS_QUERY_REGISTRY (object_reg, iDynamics);
	if (!dyn)
	{
		csReport (object_reg, CS_REPORTER_SEVERITY_ERROR,
			"crystalspace.application.phystut",
			"No iDynamics plugin!");
		return false;
	}
	
	// Get the phantom plugin.
	phantom = CS_QUERY_REGISTRY (object_reg, iPhantom);
	if (!phantom) {
		fprintf (stderr, "No csPhantom shared class!\n");
	}
	
	validConstruction = phantom->getValidConstruction();
	
	// Open the main system. This will open all the previously loaded plug-ins.
	if (!csInitializer::OpenApplication (object_reg))
	{
		csReport (object_reg, CS_REPORTER_SEVERITY_ERROR,
			"crystalspace.application.demo2",
			"Error opening system!");
		return false;
	}
	
	// Create the dynamic system.
	dynSys = dyn->CreateSystem ();
	if (dynSys == NULL)
	{
		csReport (object_reg, CS_REPORTER_SEVERITY_ERROR,
			"crystalspace.application.phystut",
			"Error creating dynamic system!");
		return false;
	}

	dynSys->SetGravity (csVector3 (0,-5,0));
	
	// Find the font we'll use
	// Load the default font name
	Font = g2d->GetFontServer ()->LoadFont (CSFONT_COURIER);
	Font->SetSize(50);

	engine->SetLightingCacheMode (0);

	Create2DImages();

	if(cmdLevel) {
		loadLevel(7);
	}

	return true;
}

/** Loads the level to play and loads the ball and the phantom cursor. Creates rigid bodies for all objects. */
bool Demo2::loadLevel(int level) {

	LoadMap(level);
	CreateRigidBodies();	
	
	//********************************* Ball ************************************************

	// Load a texture for our sprite.
	iTextureWrapper* txt1 = loader->LoadTexture ("spark","/lib/std/spark.png", CS_TEXTURE_3D, g3d->GetTextureManager (), true);
	if (txt1 == NULL)
	{
		csReport (object_reg, CS_REPORTER_SEVERITY_ERROR,
			"crystalspace.application.demo2", 
			"Error loading texture!");
		return false;
	}
	
	mw1 = engine->GetMaterialList()->FindByName("spark");

	const char* classId = "crystalspace.mesh.object.ball";
	csRef<iMeshFactoryWrapper> mesh_fact1 (engine->CreateMeshFactory(classId,"ballFact"));
	
	csVector3 meshposition1(0,0,0);
	
	csRef<iMeshWrapper> sphere = engine->CreateMeshWrapper(mesh_fact1,"sphere", room, meshposition1);
	
	csVector3 meshradius(0.5,0.5,0.5);
	csRef<iBallState> ballstate1 (SCF_QUERY_INTERFACE( sphere->GetMeshObject(),iBallState));
		
	ballstate1->SetRadius( meshradius.x, meshradius.y, meshradius.z );
	ballstate1->SetRimVertices( 20 );
	ballstate1->SetMaterialWrapper( mw1 );
	ballstate1->SetLighting(true);
	sphere->DeferUpdateLighting (CS_NLIGHT_STATIC|CS_NLIGHT_DYNAMIC, 10);
	
	// Create a body and attach the mesh.
	rbSphere = dynSys->CreateBody ();
	rbSphere->MakeDynamic();
	rbSphere->SetProperties (1, csVector3 (0,0,0), csMatrix3 ());
	rbSphere->AttachMesh (sphere);
	
	rbSphere->AttachColliderSphere (0.5, csVector3(0,0,0), 10, 1, 0.8); // 10 1 .8
	rbSphere->SetPosition (view->GetCamera ()->GetTransform ().GetOrigin());
	rigidBodies[engine->GetMeshes()->GetCount()-1] = rbSphere;
	
	//********************************* Phantom Cursor ***************************************************
	
	csRef<iMeshFactoryWrapper> mesh_fact (engine->CreateMeshFactory(classId,"ballFact"));
	csVector3 meshposition(0,0,0);
	
	ball = engine->CreateMeshWrapper(mesh_fact,"PHANTOM", room, meshposition);
	
	csVector3 phantomradius(0.05,0.05,0.05);
	csRef<iBallState> ballstate (SCF_QUERY_INTERFACE( ball->GetMeshObject(),iBallState));
	
	l_mat = (engine->CreateBaseMaterial (NULL, 0, NULL, NULL));
	l_mat->SetFlatColor (csRGBcolor (0, 255, 0));
	mw2 = engine->GetMaterialList ()->NewMaterial (l_mat);
	mw2->QueryObject ()->SetName ("color");
	mw2->Register (g3d->GetTextureManager ());
	mw2->GetMaterialHandle ()->Prepare ();
	
	ballstate->SetRadius( phantomradius.x, phantomradius.y, phantomradius.z );
	ballstate->SetRimVertices( 10 );
	ballstate->SetMaterialWrapper( mw2 );
	ballstate->SetLighting(true);
	ball->DeferUpdateLighting (CS_NLIGHT_STATIC|CS_NLIGHT_DYNAMIC, 10);
	
	engine->Prepare ();

	// Create haptic environment.
	if (validConstruction)
		phantom->setMeshList(engine->GetMeshes());

	// Start the phantom device.
	phantom->start();

	return true;
}

/** Loads the specified map and puts the camera in the correct position. */
bool Demo2::LoadMap(int level) {

	csRef<iVFS> VFS (CS_QUERY_REGISTRY (object_reg, iVFS));
	switch(level){
	case 1:
		map_dir = new char[512];
		strcpy(map_dir,"/lev/labyrinth1");
		break;
	case 2:
		map_dir = new char[512];
		strcpy(map_dir,"/lev/labyrinth2");
		break;
	case 3:
		map_dir = new char[512];
		strcpy(map_dir,"/lev/labyrinth3");
		break;
	case 4:
		map_dir = new char[512];
		strcpy(map_dir,"/lev/labyrinth4");
		break;
	case 5:
		map_dir = new char[512];
		strcpy(map_dir,"/lev/labyrinth5");
		break;
	case 6:
		map_dir = new char[512];
		strcpy(map_dir,"/lev/labyrinth6");
		break;
	default:
		break;
	}

	// Set VFS current directory to the level we want to load.
	
	VFS->ChDir (map_dir);
	// Load the level file which is called 'world'.
	if (!loader->LoadMapFile ("world"))
	{
		csReport (object_reg, CS_REPORTER_SEVERITY_ERROR,
			"crystalspace.application.simple",
			"Couldn't load level!");
		return false;
	}
	// Find the starting position in this level.
	csVector3 pos (0, CAMERA_POSITION_Y, 0);
	if (engine->GetCameraPositions ()->GetCount () > 0)
	{
		// There is a valid starting position defined in the level file.
		iCameraPosition* campos = engine->GetCameraPositions ()->Get (0);
		room = engine->GetSectors ()->FindByName (campos->GetSector ());
		pos = campos->GetPosition ();
		pos.y = CAMERA_POSITION_Y;
		if (validConstruction) phantom->setWorkspacePosition(pos.x, PHANTOM_POSITION_Y, pos.z);
	}
	else
	{
		// We didn't find a valid starting position. So by default
		// we go to room called 'room' at position (0,0,0).
		room = engine->GetSectors ()->FindByName ("room");
		if (room != NULL)
			printf("lr finnas ett room\n");
	}
	if (!room)
	{
		csReport (object_reg, CS_REPORTER_SEVERITY_ERROR,
			"crystalspace.application.simple",
			"Can't find a valid starting position!");
		return false;
	}
	view = csPtr<iView> (new csView (engine, g3d));
	view->GetCamera ()->GetTransform ().SetOrigin (pos);
	view->GetCamera ()->SetSector (room);
	view->GetCamera ()->GetTransform ().RotateThis (CS_VEC_TILT_DOWN, 1.57);
	lastPhantomPosition = pos;
	cameraStartPosition = pos;
	if (!workspaceRotated) {
		phantom->rotateWorkspace(1.57, CS_VEC_TILT_DOWN.x, CS_VEC_TILT_DOWN.y, CS_VEC_TILT_DOWN.z);
		workspaceRotated = true;
	}
	
	return true;	
}

/** Writes a text to the screen. */
void Demo2::GfxWrite (int x, int y, int fg, int bg, char *str, ...) {

  va_list arg;
  char buf[200];

  va_start (arg, str);
  vsprintf (buf, str, arg);
  va_end (arg);
  g2d->Write (Font, x, y, fg, bg, buf);
}

/** Updates the camera and the phantom position. 
*	Updates the dynamics system. 
*	Updates the phantom cursor. 
*	Keeps track of the time. */
void Demo2::updatePhantomFrame() {
	
	// Wait a while for the phantom device to start before updating the dynamics system.
	if(frames<500000) {
		frames++;
		return;
	}

	// Set start time.
	if (firstTime) {
		startTime = (float)((vc->GetCurrentTicks () / 1000.0));
		firstTime = false;
	}

	// Check if the ball has reached the goal position.
	csVector3 spherePos = rbSphere->GetPosition();
	if(sqrt(pow((spherePos.x - goalPosition.x),2) + pow((spherePos.z - goalPosition.z),2)) < 0.5) {
		stopTime = (float)((vc->GetCurrentTicks () / 1000.0)) - startTime;
		mode = 1;
	}
	
	// Check if the game is over. The ball has fallen below the ground.
	if(spherePos.y < -0.4)
		mode = 2;
	
	csTicks elapsed_time = vc->GetElapsedTicks ();
		
	// Now rotate the camera according to keyboard state
	float speed = (float)((elapsed_time / 1000.0) * (0.03 * 35));//20);
	
	iCamera* c = view->GetCamera();
	csVector3 move;
	csVector3 campos = c->GetTransform().GetOrigin();

	if (kbd->GetKeyState (CSKEY_RIGHT) && campos.x < maxX) {
		c->Move (CS_VEC_RIGHT * speed);
	}
	if (kbd->GetKeyState (CSKEY_LEFT) && campos.x > minX ) {
		c->Move (CS_VEC_LEFT * speed);
	}
	if (kbd->GetKeyState (CSKEY_PGUP)) {
		c->Move (CS_VEC_BACKWARD * speed);
		if (c->GetTransform().GetOrigin().y > MAX_CAMERA_HEIGHT)
			c->Move (CS_VEC_FORWARD * speed);
	}
	
	if (kbd->GetKeyState (CSKEY_PGDN)) {
		c->Move (CS_VEC_FORWARD * speed);
		if (c->GetTransform().GetOrigin().y < MIN_CAMERA_HEIGHT)
			c->Move (CS_VEC_BACKWARD * speed);
	}

	if (kbd->GetKeyState (CSKEY_UP) && campos.z < maxZ) {
		c->Move (CS_VEC_UP * speed);
	}
	if (kbd->GetKeyState (CSKEY_DOWN) && campos.z > minZ) {
		c->Move (CS_VEC_DOWN * speed);
	}

	// The phantom position should be updated in smaller steps.
	csVector3 pos = c->GetTransform().GetOrigin();
	pos.y = CAMERA_POSITION_Y;
	move = pos - lastPhantomPosition;
	lastPhantomPosition = pos;
	move = move/20;

	// Take small steps.
	const float maxStep = speed/20;
	float ta = 0;
	float tb = speed;
	int i = 0;
	while (ta < speed)
	{
		if (tb - ta > maxStep)
			tb = ta + maxStep;

		std::list<std::pair<int,csVector3> > list = phantom->getDynamicForceList();
		
		// Add force to rigid bodies if they have been touched by the phantom.
		while(!list.empty()){
			std::pair<int,csVector3> pair = (std::pair<int,csVector3>)list.front();
			list.pop_front();
			rigidBodies[pair.first]->AddForceAtPos(pair.second,phantom->getStylusPosition());
			
		}
		
		// Update the dynamics system.
		dyn->Step (tb-ta);
		
		ta = tb;
		tb = speed;
		i++;
		if (i<20) {
			pos += move;
			// Updates the phantom workspace position.
			phantom->setWorkspacePosition(pos.x,PHANTOM_POSITION_Y,pos.z);
		}
	}
	
	// Tell 3D driver we're going to display 3D things.
	if (!g3d->BeginDraw (engine->GetBeginDrawFlags () | CSDRAW_3DGRAPHICS))
		return;
	
	// Change color of the phantom cursor if it collides with an object.
	if(phantom->isCollision())					
		l_mat->SetFlatColor(csRGBcolor (255, 0, 50));
	else												
		l_mat->SetFlatColor(csRGBcolor (0, 255, 50));
	mw2->Register (g3d->GetTextureManager ());
	mw2->GetMaterialHandle ()->Prepare ();
	
	// Updates the position of the phantom cursor.
	ball->GetMovable()->SetPosition(phantom->getStylusPosition());
	ball->GetMovable()->UpdateMove();
	
	view->Draw ();
	g3d->FinishDraw ();
	
	// Writes the elapsed time to the screen.
	g3d->BeginDraw(CSDRAW_2DGRAPHICS);
	char buffer[200];
	sprintf (buffer, "%i",(int) (vc->GetCurrentTicks () / 1000.0 - startTime));
	GfxWrite(g2d->GetWidth() - 30, g2d->GetHeight() - 30, 16777215, 0, "%s", buffer);
	g3d->FinishDraw();
	g3d->Print (NULL);
}

/** Remove the prevoiusly played level. */
void Demo2::ClearLevel() {

	dynSys->RemoveBody(rbSphere);
	rbSphere = NULL;
	dyn->RemoveSystem (dynSys);
	dynSys = NULL;
	rigidBodies.clear();
	frames = 0;
	firstTime = true;

	// Create the dynamic system.
	dynSys = dyn->CreateSystem ();
	if (dynSys == NULL)
	{
		csReport (object_reg, CS_REPORTER_SEVERITY_ERROR,
			"crystalspace.application.phystut",
			"Error creating dynamic system!");
		return;
	}
	dynSys->SetGravity (csVector3 (0,-5,0));
	
}

/** Create rigid bodies for all geometry in the map. */
void Demo2::CreateRigidBodies() {

	csRef<iMeshList> meshList = engine->GetMeshes();
	int meshListSize = meshList->GetCount();
	std::vector<csRef<iRigidBody> > bodies(meshListSize+1);
	std::list<csRef<iRigidBody> > dynamicBodies;
	int index = 0;
	int nbrDyn = 0;
	std::list<int> dynMeshes;

	 // Check all the meshes
	for(int i = 0;i<meshListSize;i++) {
		iMeshWrapper* meshWrapper = meshList->Get(i);
		iPolygonMesh* polyMesh = meshWrapper->GetMeshObject()->GetObjectModel()->GetPolygonMeshColldet();
		// Just interrested in the ones that implements polymesh.
		if (polyMesh != NULL && strcmp(meshWrapper->QueryObject()->GetName(), "PHANTOM") != 0) { 
			// Do the following for a dynamic box, a static box or a goal.
			if(strcmp(meshWrapper->QueryObject()->GetName(), "dynamic_box") == 0 || 
				strcmp(meshWrapper->QueryObject()->GetName(), "static_box") == 0 || 
				strcmp(meshWrapper->QueryObject()->GetName(), "goal") == 0) {
				
				csBox3 boundBox;
				meshWrapper->GetWorldBoundingBox(boundBox);
				float sizeX = boundBox.MaxX()-boundBox.MinX();
				float sizeY = boundBox.MaxY()-boundBox.MinY();
				float sizeZ = boundBox.MaxZ()-boundBox.MinZ();
				
				// Store the min and max values for all geometry.
				if(boundBox.MinX() < minX)
					minX = boundBox.MinX();
				
				if(boundBox.MaxX() > maxX)
					maxX = boundBox.MaxX();
				
				if(boundBox.MinZ() < minZ)
					minZ = boundBox.MinZ();
				
				if(boundBox.MaxZ() > maxZ)
					maxZ = boundBox.MaxZ();
				
				// Dynamic box.
				if(strcmp(meshWrapper->QueryObject()->GetName(), "dynamic_box") == 0) {
					iTextureManager* txtmgr = g3d->GetTextureManager();
					iTextureWrapper* txt = loader->LoadTexture ("WOOD",
						"/lib/std/WOOD.png", CS_TEXTURE_3D, txtmgr, true);
					if (txt == NULL)
					{
						csReport (object_reg, CS_REPORTER_SEVERITY_ERROR,
							"crystalspace.application.simple2",
							"Error loading texture!");
						Sleep(2000);
						exit(0);
					}
					
					// Load a sprite template from disk.
					csRef<iMeshFactoryWrapper> imeshfact(loader->LoadMeshObjectFactory("/lib/std/sprite1"));
					if (imeshfact == NULL)
					{
						csReport (object_reg, CS_REPORTER_SEVERITY_ERROR,
							"crystalspace.application.demo2",
							"Error loading mesh object factory!");
						Sleep(2000);
						exit(0);
					}

					// Adjust the size of the sprite.
					csMatrix3 m(1.0/sizeX,0,0,0,1.0/sizeY,0,0,0,1.0/sizeZ);
					csReversibleTransform trans = csReversibleTransform (m, csVector3 (0));
					imeshfact->HardTransform (trans);
					
					// Create the sprite and add it to the engine.
					csRef<iMeshWrapper> newMeshWrapper(engine->CreateMeshWrapper(imeshfact, meshWrapper->QueryObject()->GetName(), room,csVector3 (0,0,0)));
					csRef<iSprite3DState> spstate1 (
						SCF_QUERY_INTERFACE (newMeshWrapper->GetMeshObject (), iSprite3DState));
					spstate1->SetAction ("default");

					// Create a rigid body.
					csRef<iRigidBody> rbody = dynSys->CreateBody ();
					rbody->SetProperties (0.3, csVector3 (0), csMatrix3 ());
					rbody->AttachMesh (newMeshWrapper);
					rbody->MakeDynamic();
					
					// Create and attach a box collider.
					const csMatrix3 tm;
					const csVector3 tv (0);
					csOrthoTransform t (tm, tv);
					csVector3 size (sizeX, sizeY, sizeZ); // This should be the same size as the mesh.
					rbody->AttachColliderBox (size, t, 2, 1, 0.9);
					rbody->SetPosition (boundBox.GetCenter());
					dynamicBodies.push_back(rbody);
					
					// Store which mesh that should be removed from the meshlist.
					dynMeshes.push_back(i);
					nbrDyn++;
					
					index--;
				}
				// Static box and goal.
				else {

					// Create a body and attach the mesh.
					bodies[index] = dynSys->CreateBody ();
					bodies[index]->SetProperties (0.3, csVector3 (0), csMatrix3 ());
					bodies[index]->AttachMesh (meshWrapper);
					bodies[index]->MakeStatic();
					
					// Create and attach a box collider.
					const csMatrix3 tm;
					const csVector3 tv (0);
					csOrthoTransform t (tm, tv);
					csVector3 size (sizeX, sizeY, sizeZ); // This should be the same size as the mesh.
					bodies[index]->AttachColliderBox (size, t, 2, 1, 0.9);
					bodies[index]->SetPosition (boundBox.GetCenter());
					if (strcmp(meshWrapper->QueryObject()->GetName(), "goal") == 0)
						goalPosition = boundBox.GetCenter();
				}
			}
		}
		index++;
	}

	// Copy the dynamic bodies.
	int count = 0;
	std::vector<std::pair<int,csVector3> > startPos(nbrDyn);
	while(!dynamicBodies.empty()){
		csRef <iRigidBody> body = dynamicBodies.front();
		dynamicBodies.pop_front();
		bodies[index] = body;
		std::pair<int,csVector3> startPair(index,bodies[index]->GetPosition());
		startPos[count] = startPair;
		index++;
		count++;
	}
	// Remove the old mesh objects from the mesh list.
	while(!dynMeshes.empty()){
		meshList->Remove(dynMeshes.back());
		dynMeshes.pop_back();
	}

	rigidBodies = bodies;
	startPositions = startPos;

	// Add an offset for the camera boundaries.
	minX += OFFSET;
	maxX -= OFFSET;
	minZ += OFFSET;
	maxZ -= OFFSET;
}

/** Check mode to show the correct frame. */
void Demo2::SetupFrame () {

	switch(mode){
	case 0:
		startFrame();
		break;
	case 1:
		gameWonFrame();
		phantom->stop();
		ClearLevel();
		break;
	case 2:
		gameOverFrame();
		phantom->stop();
		ClearLevel();
		break;
	default:
		if (validConstruction)
			updatePhantomFrame();
	}
}

/** Draw the startup image. */
void Demo2::startFrame() {

	g3d->BeginDraw (engine->GetBeginDrawFlags () | CSDRAW_3DGRAPHICS);
	startScreen->DrawScaled (g3d, 0 , 0, 640, 480);
	g3d->FinishDraw();
	g3d->Print (NULL);
}

/** Draw the congratulations image. */
void Demo2::gameWonFrame() {

	g3d->BeginDraw (engine->GetBeginDrawFlags () | CSDRAW_3DGRAPHICS);
	gameWonScreen->DrawScaled (g3d, 0 , 0, 640, 480);
	char buffer[200];
	sprintf (buffer, "You finished this level in %i seconds",(int) stopTime);
	GfxWrite(185, g2d->GetHeight() - 240, -1, 3026514, "%s", buffer);
	g3d->FinishDraw();
	g3d->Print (NULL);
}

/** Draw the game over image. */
void Demo2::gameOverFrame() {

	g3d->BeginDraw (engine->GetBeginDrawFlags () | CSDRAW_3DGRAPHICS);
	gameOverScreen->DrawScaled (g3d, 0 , 0, 640, 480);
	g3d->FinishDraw();
	g3d->Print (NULL);
}

/** Finish drawing the frame. Do nothing. */
void Demo2::FinishFrame () {
}

/** Handles a system or a keyboard event.*/
bool Demo2::HandleEvent (iEvent& ev) {

	if (ev.Type == csevBroadcast && ev.Command.Code == cscmdProcess)
	{
		SetupFrame ();
		return true;
	}
	else if (ev.Type == csevBroadcast && ev.Command.Code == cscmdFinalProcess)
	{
		FinishFrame ();
		return true;
	}
	// Quit the application.
	else if (ev.Type == csevKeyDown && ev.Key.Code == 'q')
	{
		csRef<iEventQueue> q (CS_QUERY_REGISTRY (object_reg, iEventQueue));
		pt->phantom->stop();
		if (q) q->GetEventOutlet()->Broadcast (cscmdQuit);
		return true;
	}
	// Choose level.
	else if (mode == 0) {
		if(ev.Type == csevKeyDown && ev.Key.Code == '1') {
			loadLevel(1);
			mode = 3;
		}
		else if(ev.Type == csevKeyDown && ev.Key.Code == '2') {
			loadLevel(2);
			mode = 3;
		}
		else if(ev.Type == csevKeyDown && ev.Key.Code == '3') {
			loadLevel(3);
			mode = 3;
		}
		else if(ev.Type == csevKeyDown && ev.Key.Code == '4') {
			loadLevel(4);
			mode = 3;
		}
		else if(ev.Type == csevKeyDown && ev.Key.Code == '5') {
			loadLevel(5);
			mode = 3;
		}
		else if(ev.Type == csevKeyDown && ev.Key.Code == '6') {
			loadLevel(6);
			mode = 3;
		}
	}
	// Press space to continue from congratulations screen to startup screen.
	else if (mode == 1 && ev.Type == csevKeyDown && ev.Key.Code == CSKEY_SPACE) {
		mode = 0;
		phantom->stop();
		ClearLevel();
	}
	// Press space to continue from game over screen to startup screen.
	else if (mode == 2 && ev.Type == csevKeyDown && ev.Key.Code == CSKEY_SPACE) {
		mode = 0;
		phantom->stop();
		ClearLevel();
	}
	// Playing.
	else if (mode == 3) {
		// Restart the game at the start position.
		if (ev.Type == csevKeyDown && ev.Key.Code == 'r')
		{
			phantom->setForcesOn(false);
			for(int i=0;i<startPositions.size();i++) {
				rigidBodies[startPositions[i].first]->SetPosition(startPositions[i].second);
				const csMatrix3 tm;
				const csVector3 tv (0);
				csOrthoTransform t (tm, tv);
				rigidBodies[startPositions[i].first]->SetOrientation(tm);
				rigidBodies[startPositions[i].first]->SetLinearVelocity(csVector3(0,0,0));
				rigidBodies[startPositions[i].first]->SetAngularVelocity(csVector3(0,0,0));
			}
			view->GetCamera()->GetTransform().SetOrigin(cameraStartPosition);
			rbSphere->SetPosition(cameraStartPosition);
			rbSphere->SetLinearVelocity(csVector3(0,0,0));
			rbSphere->SetAngularVelocity(csVector3(0,0,0));
			phantom->setWorkspacePosition(cameraStartPosition.x,PHANTOM_POSITION_Y, cameraStartPosition.z);
			phantom->setForcesOn(true);
		}
		
		// Go back to the startup screen.
		else if(ev.Type == csevKeyDown && ev.Key.Code == CSKEY_ESC)
		{
			mode = 0;
			phantom->stop();
			ClearLevel();
		}
	}
							
	return false;
}

/** Handle an event. */
bool Demo2::SimpleEventHandler (iEvent& ev) {

	return pt->HandleEvent (ev);
}

/** Start the default run loop */
void Demo2::Start () {

	csDefaultRunLoop (object_reg);
}

/** Creates the 2D images used for the different 2D screens. */
void Demo2::Create2DImages(void) {
	
	int w, h;
	iTextureWrapper *texh;
	iTextureHandle* phTex;
	
	texh = loader->LoadTexture ("startup_screen",
		"/lib/std/startup_screen.png", CS_TEXTURE_3D, g3d->GetTextureManager (), true);
	if (texh == NULL)
	{
		csReport (object_reg, CS_REPORTER_SEVERITY_ERROR,
			"crystalspace.application.demo2",
			"Error loading texture!");
		return ;
	}
	
	if (texh)
	{
		phTex = texh->GetTextureHandle();
		if (phTex)
		{
			phTex->GetMipMapDimensions (0, w, h);
			startScreen = new csSimplePixmap (phTex, 0, 0, w, h);
		}
		
	}
	
	texh = loader->LoadTexture ("gamewon_screen",
		"/lib/std/gamewon_screen.png", CS_TEXTURE_3D, g3d->GetTextureManager (), true);
	if (texh == NULL)
	{
		csReport (object_reg, CS_REPORTER_SEVERITY_ERROR,
			"crystalspace.application.demo2",
			"Error loading texture!");
		return ;
	}
	
	if (texh)
	{
		phTex = texh->GetTextureHandle();
		if (phTex)
		{
			phTex->GetMipMapDimensions (0, w, h);
			gameWonScreen = new csSimplePixmap (phTex, 0, 0, w, h);
		}	  
	}
    texh = loader->LoadTexture ("gameover_screen",
		"/lib/std/gameover_screen.png", CS_TEXTURE_3D, g3d->GetTextureManager (), true);
	if (texh == NULL)
	{
		csReport (object_reg, CS_REPORTER_SEVERITY_ERROR,
			"crystalspace.application.demo2",
			"Error loading texture!");
		return ;
	}
	
	if (texh)
	{
		phTex = texh->GetTextureHandle();
		if (phTex)
		{
			phTex->GetMipMapDimensions (0, w, h);
			gameOverScreen = new csSimplePixmap (phTex, 0, 0, w, h);
		}	  
	}
}

// Main method.
int main(int argc, char *argv[]) {
	
	iObjectRegistry* object_reg = csInitializer::CreateEnvironment (argc, argv);
	pt = new Demo2(object_reg); 
	
	pt->Initialize();
	pt->Start();
	
	csInitializer::DestroyApplication (object_reg);

	return 0;	
}